#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include "config.h"


char *lastname(path)
    char *path;
{
    path = strrchr(path, '/');
    if (!path)
	return NULL;
    else
	return ++path;
}


char *show_pathname(path)
    char *path;
{
    /*
     * Remove possible leading HOMEDIR
     */
    if (strncmp(path, HOMEDIR, sizeof(HOMEDIR)-1) == 0)
	path += sizeof(HOMEDIR)-1;

    if (*path == '\0')
	strcpy(path, "/");
    
    return path;
}


char *fix_pathname(path)
    char *path;
{
    extern char *getwd();
    static char buf[2048];
    char *cp, *top, *point;

    
    if (!path)
	cp = getwd(buf);
    else
    {
	if (*path != '/')
	    cp = getwd(buf); /* Relative path? Start with current directory */
	else
	{
	    /* Absolute path (from HOMEDIR) */
	    strcpy(buf, HOMEDIR);
	    cp = buf;
	    ++path;
	}

	while (*cp)
	    ++cp;

	strcat(cp, "/");
	strcat(cp, path);

	/* Point "top" at the "/" before "path" */
	top = buf + sizeof(HOMEDIR)-1;
	++cp;
	
	/*
	 * Ok, we now have a full path like "HOMEDIR/path", now fix it so
	 * that it doesn't points at files outside of HOMEDIR.
	 */
	while (*cp)
	{
	    /*
	     * At start of pathname component
	     */
	    if (cp[0] == '.' && cp[1] == '.' &&
		(cp[2] == '/' || cp[2] == '\0'))
	    {
		/*
		 * Got a "..", back up one component
		 */
		point = cp+2;
		
		--cp; /* Go up one char, now pointing at a "/" */

		/*
		 * We've backed to far - that's illegal
		 */
		if (cp <= top)
		    return NULL;
		
		while (--cp > top && *cp != '/')
		    ;

		++cp; /* Point at first char after the "/" */
		
		strcpy(cp, point);
		if (!*cp)
		    break;
	    }
	    else if (*cp == '/')
	    {
		/*
		 * Got a "//" entry, remove the double "/"
		 */
		strcpy(cp, cp+1);
	    }
	    else
	    {
		/*
		 * Normal path component, just skip it
		 */
		while (*cp && *cp != '/')
		    ++cp;
		if (*cp)
		    ++cp;
	    }
	}
    }

    if (cp > buf && cp[-1] == '/')
	cp[-1] = '\0';

    return buf;
}


FILE *open_mail(return_address, subject)
  char *return_address;
  char *subject;
{
  char buf[256];
  FILE *fp;
  
  if (strlen(return_address) > 200)
    exit(1);
  
  sprintf(buf, SEND_EMAIL, return_address);
  fp = popen(buf, "w");
  if (!fp)
  {
    syslog(LOG_ERR, "Unable to popen(\"%s\", \"w\"): %m", buf);
    exit(2);
  }

  fprintf(fp, "To: %s\n", return_address);
  fprintf(fp, "Errors-To: %s\n", ERRORUSER);
  fprintf(fp, "Reply-To: ftpserv@lysator.liu.se\n");
  fprintf(fp, "Subject: %s\n", subject);
  putc('\n', fp);

  return fp;
}


send_fp(ofp, ifp)
  FILE *ofp;
  FILE *ifp;
{
  int lines = 0;
  char buf[256];
  
  while (fgets(buf, sizeof(buf)-1, ifp))
  {
    lines++;
    fputs(buf, ofp);
  }

  return lines;
}


send_file(ofp, path)
FILE *ofp;
char *path;
{
  FILE *ifp;

  ifp = fopen(path, "r");
  if (!ifp)
    return -1;

  send_fp(ofp, ifp);
  
  fclose(ifp);
}

main(argc,argv)
  int argc;
  char *argv[];
{
  FILE *fp, *fp2;
  char buf[256];
  char *cp, *cp2, *path, *fname, *bc;
  char *return_address = NULL;
  char *subject = "Foo";


  if (fork())
    _exit(0);
  
  signal(SIGPIPE, SIG_IGN);

  chdir(HOMEDIR);
  
  /* Skip until body of message */
  while (fgets(buf, sizeof(buf)-1, stdin))
  {
    for (cp = buf; *cp && isspace(*cp); cp++)
      ;
    if (*cp == '\0')
      break;

    cp = strtok(buf, ":");
    if (cp)
    {
      cp2 = strtok(NULL, "\n");
      while (cp2 && *cp2 && isspace(*cp2))
	cp2++;
      
      if (stricmp(cp, "Reply-To") == 0)
	return_address = strdup(cp2);
      else if (stricmp(cp, "From") == 0 && return_address == NULL)
	return_address = strdup(cp2);
      else if (stricmp(cp, "Subject") == 0)
	subject = strdup(cp2);
    }
  }

  openlog("ftpserv", LOG_PID, LOG_LOCAL2);
  syslog(LOG_NOTICE, "CONN %s", return_address);
	 
  if (!return_address || strstr(return_address, "MAILER-DAEMON") ||
      strstr(return_address, "mailer-daemon") ||
      strstr(return_address, "Mailer-Daemon"))
  {
      exit(0);
  }

  /* Parse body of message */
  while (fgets(buf, sizeof(buf)-1, stdin))
  {
    bc = strdup(buf);
    if (bc[strlen(bc)-1] == '\n')
      bc[strlen(bc)-1] = '\0';
    
    cp = strtok(buf, " \t\n");
    if (!cp)
      continue;
    cp2 = strtok(NULL, " \t\n");
    if (cp2)
      cp2 = strdup(cp2);
    while (cp2 && *cp2 && isspace(*cp2))
      cp2++;
    
    if (stricmp(cp, "help") == 0)
    {
      int ecode;
      
      fp = open_mail(return_address, bc);
      send_file(fp, HELPFILE);
      ecode = pclose(fp);
      if (ecode)
	syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
    }
    else if (stricmp(cp, "dir") == 0)
    {
      int ecode;
      
      syslog(LOG_DEBUG, "DIR %s", cp2 ? cp2 : "");
      fp = open_mail(return_address, bc);
    
      path = fix_pathname(cp2);
      if (!path)
      {
	fprintf(fp, "Illegal pathname in DIR command: %s\n", cp2);
	exit(0);
      }
      else
      {
	fprintf(fp, "DIR listing of '%s':\n", show_pathname(path));
	
	sprintf(buf,"/bin/ls -Flg %s", path);
	fp2 = popen(buf, "r");
	send_fp(fp, fp2);
      }
      ecode = pclose(fp);
      if (ecode)
	syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
    }
    else if (stricmp(cp, "uusend") == 0)
    {
	int ecode;
	char buf2[2000];
	FILE *fp3;
	int lines, index;
	int part, parts;

	
	if (!cp2 || !(path = fix_pathname(cp2)) || !(fname = lastname(path)))
	{
	    fp = open_mail(return_address, bc);
	    fprintf(fp, "Illegal pathname in UUSEND command: %s\n", cp2);
	    ecode = pclose(fp);
	    if (ecode)
		syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	    exit(0);
	}

	strcpy(buf, "/tmp/ftpserv.XXXXXX");
	mktemp(buf);
		
	sprintf(buf2,"uuencode %s %s.uu", path, fname);
	fp2 = popen(buf2, "r");
	if (!fp2)
	{
	    syslog(LOG_ERR, "popen() failed: %m");
	    exit(0);
	}
	fp3 = fopen(buf, "w");
	if (!fp3)
	    exit(0);
	lines = send_fp(fp3, fp2);
	fclose(fp3);
	ecode = pclose(fp2);
	if (ecode)
	{
	    syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	    exit(0);
	}
	
	fp3 = fopen(buf, "r");
	if (!fp3)
	{
	    syslog(LOG_ERR, "fopen() returned: %m");
	    exit(0);
	}
	
	parts = (lines / PARTLINESIZE) + 1;
	syslog(LOG_NOTICE, "UUSEND %s (%d parts)", path, parts);
	for (part = 1; part <= parts; part++)
	{
	    sprintf(buf2, "%s (Part %d/%d)", bc, part, parts);
	    fp2 = open_mail(return_address, buf2);
	    for (index = 0;
		 index < PARTLINESIZE && fgets(buf2, sizeof(buf2)-1, fp3);
		 index++)
		fputs(buf2, fp2);
	    ecode = pclose(fp2);
	    if (ecode)
		syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	}
	fclose(fp3);
	unlink(buf);
    }
    else if (stricmp(cp, "send") == 0)
    {
      int ecode;

      
      fp = open_mail(return_address, bc);
      if (!cp2 || !(path = fix_pathname(cp2)))
      {
	  fprintf(fp, "Illegal pathname in SEND command: %s\n", cp2);
	  ecode = pclose(fp);
	  if (ecode)
	    syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	  exit(0);
      }
      else
      {
	  if (send_file(fp, path) < 0)
	  {
	    fprintf(fp, "Send text file failed for: %s\n", cp2);
	    ecode = pclose(fp);
	    if (ecode)
	      syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	    exit(0);
	  }
	  syslog(LOG_NOTICE, "SEND %s", cp2);
      }
      ecode = pclose(fp);
      if (ecode)
	syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
    }
    else if (stricmp(cp, "cd") == 0)
    {
      int ecode;
      
      syslog(LOG_DEBUG, "CD %s", cp2 ? cp2 : "");
      if (!cp2)
	chdir(HOMEDIR);
      else
      {
	  if (!(path = fix_pathname(cp2)))
	  {
	      fp = open_mail(return_address, bc);
	      fprintf(fp, "Illegal pathname in CD command: %s\n", cp2);
	      ecode = pclose(fp);
	      if (ecode)
		  syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	      exit(0);
	  }
	  else
	  {
	  if (chdir(path) == -1)
	  {
	    fp = open_mail(return_address, bc);
	    fprintf(fp, "Cannot chdir to directory: %s\n", cp2);
	    ecode = pclose(fp);
	    if (ecode)
	      syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
	    exit(0);
	  }
	}
      }
    }
    else if (stricmp(cp, "quit") == 0)
      exit(0);
    else
    {
      int ecode;
      
      fp = open_mail(return_address, bc);
      fprintf(fp, "Unknown command: %s\n\n", cp);
      ecode = pclose(fp);
      if (ecode)
	syslog(LOG_ERR, "pclose() returned: %d: %m", ecode);
      exit(0);
    }
  }

  exit(0);
}
